
--[[

Retro Browser - Display the SB2/SB3 display on jive

--]]

-- stuff we use
local tostring, assert, type, pairs = tostring, assert, type, pairs

local oo               = require("loop.simple")
local table            = require("table")
local string           = require("string")
local mime             = require("mime")

local min              = math.min
local floor            = math.floor

local Applet           = require("jive.Applet")

local Window           = require("jive.ui.Window")
local Surface          = require("jive.ui.Surface")
local Icon             = require("jive.ui.Icon")
local Framework        = require("jive.ui.Framework")
local Widget           = require("jive.ui.Widget")

local SocketHttp       = require("jive.net.SocketHttp")

local AppletManager    = require("jive.AppletManager")
local SlimServer       = require("jive.slim.SlimServer")
local jiveMain         = jiveMain

require("applets.RetroBrowser.retro")
local drawBitsH        = retro.drawBitsH
local drawBitsV        = retro.drawBitsV

local blitClip         = Surface.blitClip
local blit             = Surface.blit
local addAnimation     = Widget.addAnimation
local removeAnimation  = Widget.removeAnimation

local debug            = require("jive.utils.debug")
local log              = require("jive.utils.log").logger("applet.retro")

local EVENT_SCROLL     = jive.ui.EVENT_SCROLL
local EVENT_KEY_ALL    = jive.ui.EVENT_KEY_ALL
local EVENT_KEY_DOWN   = jive.ui.EVENT_KEY_DOWN
local EVENT_KEY_UP     = jive.ui.EVENT_KEY_UP
local EVENT_KEY_PRESS  = jive.ui.EVENT_KEY_PRESS
local EVENT_KEY_HOLD   = jive.ui.EVENT_KEY_HOLD
local EVENT_CONSUME    = jive.ui.EVENT_CONSUME
local EVENT_UNUSED     = jive.ui.EVENT_UNUSED
local EVENT_ACTION     = jive.ui.EVENT_ACTION
local EVENT_MOTION     = 0x800000

local KEY_GO           = jive.ui.KEY_GO
local KEY_FWD          = jive.ui.KEY_FWD
local KEY_REW          = jive.ui.KEY_REW
local KEY_HOME         = jive.ui.KEY_HOME
local KEY_PLAY         = jive.ui.KEY_PLAY
local KEY_ADD          = jive.ui.KEY_ADD
local KEY_BACK         = jive.ui.KEY_BACK
local KEY_PAUSE        = jive.ui.KEY_PAUSE
local KEY_VOLUME_DOWN  = jive.ui.KEY_VOLUME_DOWN
local KEY_VOLUME_UP    = jive.ui.KEY_VOLUME_UP

local FRAME_RATE       = jive.ui.FRAME_RATE


module(...)
oo.class(_M, Applet)


local color1 = 0x10FFFFFF      -- color of pixels drawn
local color2 = 0x000000FF      -- color of background

local buttonInt = 150          -- rate limit to send emulated button presses at


-- openPlayer
function openPlayer (self, menuItem)
	local sd = AppletManager:getAppletInstance("SlimDiscovery")
	local player
	if sd then
		player = sd:getCurrentPlayer()
	end
	
	if player == nil then return end

	log:debug("RetroBrowser:openPlayer(", tostring(player), ")")

	self.player = player

	self.callback = function(chunk, err)
						if err then
							log:debug(err)
						elseif chunk then
							_process_displaystatus(self, chunk)
						end
					end

	self.oldcallbacks = {}
	self.subid = '/slim/displaystatus/' .. self.player.id

	-- store the old displaystatus callback and unsubscribe them
	for k,v in pairs(self.player.slimServer.comet.notify[self.subid]) do
		self.oldcallback = v
		self.player.slimServer.comet:unsubscribe(self.subid, v)
	end

	-- create all surfaces we are going to use - for both H and V modes
	self.window = Window("retro")

	self.sbIcon = Icon("sbIcon")
	self.display= Icon("display")

	self.sbH   = Surface:loadImage("applets/RetroBrowser/SB3H.png")
	self.sbV   = Surface:loadImage("applets/RetroBrowser/SB3V.png")

	self:horiz() 	-- start horizontal

	self.window:addWidget(self.sbIcon)
	self.window:addWidget(self.display)

	self.display:addAnimation(function() animate(self) end, FRAME_RATE)

	self.window:addListener(EVENT_SCROLL | EVENT_KEY_PRESS | EVENT_KEY_HOLD | EVENT_MOTION,
		function(event)
			return _event_handler(self, event)
		end
	)

	self:tieAndShowWindow(self.window)
end


-- layout for horizonal mode
function horiz(self)
	self.width  = 240
	self.height = 64

	-- subscribe to display updates with width 240
	self.player.slimServer.comet:subscribe(
		self.subid,
		self.callback,							  
		self.player.id,
		{ 'displaystatus', 'subscribe:bits', 'width:240' },
		true
	)

	self.dSrf = Surface:newRGB(self.width, self.height)
	self.nSrf = Surface:newRGB(self.width, self.height)
	self.oSrf = Surface:newRGB(self.width, self.height)

	self.sbIcon:setValue(self.sbH)
	self.sbIcon:setPosition(0, 62)

	self.display:setValue(self.dSrf)
	self.display:setPosition(0, 80)

	self.anim = nil
	self.oldbits = nil
	self.mode = 'H'
	self.drawBits = drawBitsH
end


-- layout for vertical mode
function vert(self)
	self.width  = 64
	self.height = 320

	-- subscribe to display updates with width 320
	self.player.slimServer.comet:subscribe(
		self.subid,
		self.callback,							  
		self.player.id,
		{ 'displaystatus', 'subscribe:bits', 'width:320' },
		true
	)

	self.dSrf = Surface:newRGB(self.width, self.height)
	self.nSrf = Surface:newRGB(self.width, self.height)
	self.oSrf = Surface:newRGB(self.width, self.height)

	self.sbIcon:setValue(self.sbV)
	self.sbIcon:setPosition(28,0)

	self.display:setValue(self.dSrf)
	self.display:setPosition(124, 0)

	self.anim = nil
	self.oldbits = nil
	self.mode = 'V'
	self.drawBits = drawBitsV
end


function _event_handler(self, event) 

	local type = event:getType()
	local button, hold, single, ret

	if type == EVENT_SCROLL then
		
		if event:getScroll() > 0 then
			button = "arrow_down"
		else
			button = "arrow_up"
		end
		
		ret = EVENT_CONSUME

	elseif type == EVENT_KEY_PRESS or type == EVENT_KEY_HOLD then

		local key = event:getKeycode()
		
		if type == EVENT_KEY_HOLD then
			hold = true
		end

		if self.mode == 'H' then
			if     key == KEY_GO then
				button = "arrow_right"
			elseif key == KEY_BACK then
				button = "arrow_left"
			elseif key == KEY_PLAY then
				button = "play"
			elseif key == KEY_ADD then
				button = "add"
			elseif key == KEY_PAUSE then
				button = "pause"
			--elseif key == KEY_HOME and hold then
			--	button = "power"
			--	hold   = nil
			end
		else
			if     key == KEY_GO then
				button = "arrow_right"
			elseif key == KEY_ADD then
				button = "arrow_left"
			elseif key == KEY_HOME then
				button = "play"
			elseif key == KEY_PLAY then
				button = "add"
			elseif key == KEY_PAUSE then
				button = "pause"
			--elseif key == KEY_BACK and hold then
			--	button = "power"
			--	hold   = nil
			end
		end

		--[[
		-- used for debugging
		if key == KEY_FWD then
			if self.mode == 'H' then
				self:vert()
			else
				self:horiz()
			end
			self.drawBits(self.nSrf, self.nSrf, self.bits, nil, color1, color2)
			blit(self.nSrf, self.dSrf, 0, 0)
			self.step = nil
			Framework:reDraw(nil)
			ret = EVEN_CONSUME
		end
		--]]

	elseif type == EVENT_MOTION then

		local x, _, _ = event:getMotion()
		local update = false

		if x > 35 and self.mode == 'H' then
			self:vert()
			update = true
		elseif x < 15 and self.mode == 'V' then
			self:horiz()
			update = true
		end
		if update then
			self.drawBits(self.nSrf, self.nSrf, self.bits, nil, color1, color2)
			blit(self.nSrf, self.dSrf, 0, 0)
			self.step = nil
			Framework:reDraw(nil)
		end

		ret = EVENT_CONSUME
	end
	
	if button then
		local ticks = Framework.getTicks()
		
		if self.buttonsent and (ticks - self.buttonsent) < buttonInt then
			log:info("suppressing button")
		else
			log:info("sending command: button ", button)
			if hold then
				self.player:send({ 'button', button .. '.hold' })
			else
				self.player:send({ 'button', button })
				if button == 'play' or button == 'add' then
					self.player:send({ 'button', button .. '.single' })
				end
			end
			self.buttonsent = ticks
		end

		ret = EVENT_CONSUME
	end
	
	return ret or EVENT_UNUSED
end


function _process_displaystatus(self, response)

	-- received bitmap display put it on the screen..
	if response.data.bits then
		self.ext     = response.data.ext
		self.bits    = mime.unb64("", response.data.bits)

		local type   = response.data.type
		local horiz  = (self.mode == 'H')
		local transition, bump

		if     type == "animate-L" then
			transition = horiz and bumpLeftH or bumpLeftV 	bump = true
		elseif type == "animate-R" then
			transition = horiz and bumpRightH or bumpRightV bump = true
		elseif type == "animate-U" then
			transition = horiz and bumpUpH or bumpUpV       bump = true
		elseif type == "animate-D" then
			transition = horiz and bumpDownH or bumpDownV   bump = true
		elseif type == "animate-r" then
			transition = horiz and pushLeftH or pushLeftV
		elseif type == "animate-l" then
			transition = horiz and pushRightH or pushRightV
		elseif type == "animate-d" then
			transition = horiz and pushUpH or pushUpV
		elseif type == "animate-u" then
			transition = horiz and pushDownH or pushDownV
		end

		if transition == nil then
			-- no animation - draw straight to new surface & copy to display surface
			self.drawBits(self.nSrf, self.nSrf, self.bits, self.oldbits, color1, color2)
			blit(self.nSrf, self.dSrf, 0, 0)
			Framework:reDraw(nil)
			self.step = nil
		else
			-- animation - flip old and new surfaces, draw onto new surface
			self.nSrf, self.oSrf = self.oSrf, self.nSrf
			self.drawBits(self.nSrf, self.oSrf, self.bits, self.oldbits, color1, color2)
			-- set up animation if different from last
			if self.trans ~= transition then
				transition(self)
				self.trans = transition
			end
			-- clear old surface if bump as we fill in with the old surface
			if bump then
				Surface.filledRectangle(self.oSrf, 0, 0, self.width, self.height, color2)
			end
			self.step = 1
		end

		self.oldbits = bits
	end
end


function animate(self)
	if self.step == nil then return end

	local thisstep = self.steps[self.step]

	if thisstep == nil then
		blit(self.nSrf, self.dSrf, 0, 0)
		self.step = nil
	else
		if thisstep.o then
			local t = thisstep.o
			blitClip(self.oSrf, t.x, t.y, t.w, t.h, self.dSrf, t.dx, t.dy)
		end
		if thisstep.n then
			local t = thisstep.n
			blitClip(self.nSrf, t.x, t.y, t.w, t.h, self.dSrf, t.dx, t.dy)
		end
		if self.static then
			local t = self.static
			blitClip(self.nSrf, t.x, t.y, t.w, t.h, self.dSrf, t.dx, t.dy)
		end
		self.step = self.step + 1
	end

	Framework:reDraw(nil)
end


function pushLeftH(self)
	self.static = nil
	self.steps  = {}
	for i = 0, self.width, floor(self.width / FRAME_RATE * 4) do
		self.steps[#self.steps+1] = {
			o = { x = i, y = 0, w = self.width - i, h = self.height, dx = 0,              dy = 0 },
			n = { x = 0, y = 0, w = i,              h = self.height, dx = self.width - i, dy = 0 },
		}
	end
end


function pushLeftV(self)
	self.static = nil
	self.steps  = {}
	for i = 0, self.height, floor(self.height / FRAME_RATE * 4) do
		self.steps[#self.steps+1] = {
			o = { x = 0, y = i, w = self.width,  h = self.height - i, dx = 0, dy = 0               },
			n = { x = 0, y = 0, w = self.height, h = i,               dx = 0, dy = self.height - i },
		}
	end
end


function pushRightH(self)
	self.static = nil
	self.steps  = {}
	for i = 0, self.width, floor(self.width / FRAME_RATE * 4) do
		self.steps[#self.steps+1] = {
			o = { x = 0,              y = 0, w = self.width - i, h = self.height, dx = i, dy = 0 },
			n = { x = self.width - i, y = 0, w = i,              h = self.height, dx = 0, dy = 0 },
		}
	end
end


function pushRightV(self)
	self.static = nil
	self.steps  = {}
	for i = 0, self.height, floor(self.height / FRAME_RATE * 4) do
		self.steps[#self.steps+1] = {
			o = { x = 0, y = 0,               w = self.width, h = self.height - i, dx = 0, dy = i },
			n = { x = 0, y = self.height - i, w = self.width, h = i,               dx = 0, dy = 0 },
		}
	end
end


function bumpLeftH(self)
	self.static = nil
	self.steps  = {
		{ n = { x = 0,              y = 0, w = self.width, h = self.height, dx = 8, dy = 0 },
		  o = { x = self.width - 8, y = 0, w = 8,          h = self.height, dx = 0, dy = 0 }, },
		{ n = { x = 0,              y = 0, w = 0,          h = self.height, dx = 0, dy = 0 }, },
	}
end


function bumpLeftV(self)
	self.static = nil
	self.steps  = {
		{ n = { x = 0, y = 0,               w = self.width, h = self.height, dx = 0, dy = 8 },
		  o = { x = 0, y = self.height - 8, w = self.width, h = 8,           dx = 0, dy = 0 }, },
		{ n = { x = 0, y = 0,               w = 0,          h = self.height, dx = 0, dy = 0 }, },
	}
end


function bumpRightH(self)
	self.static = nil
	self.steps  = {
		{ n = { x = 8, y = 0, w = self.width, h = self.height, dx = 0,              dy = 0 }, 
		  o = { x = 0, y = 0, w = 8,          h = self.height, dx = self.width - 8, dy = 0 }, },
		{ n = { x = 0, y = 0, w = 0,          h = self.height, dx = 0,              dy = 0 }, },
	}
end


function bumpRightV(self)
	self.static = nil
	self.steps  = {
		{ n = { x = 0, y = 8, w = self.width, h = self.height, dx = 0, dy = 0      },
		  o = { x = 0, y = 0, w = 8, h = self.height, dx = 0, dy = self.height - 8 }, },
		{ n = { x = 0, y = 0, w = 0, h = self.height, dx = 0, dy = 0               }, },
	}
end


function pushDownH(self)
	local top = self.height - self.ext * 2
	self.steps  = {}
	self.static = { x = 0, y = 0, w = self.width, h = top, dx = 0, dy = 0 }
	for i = 0, self.height - top, floor((self.height - top)/4) do
		self.steps[#self.steps+1] = {
			o = { x = 0, y = top,             w = self.width, h = self.height - top - i, dx = 0, dy = top + i },
			n = { x = 0, y = self.height - i, w = self.width, h = i,                     dx = 0, dy = top     },
		}
	end
end


function pushUpH(self)
	local top = self.height - self.ext * 2
	self.steps  = {}
	self.static = { x = 0, y = 0, w = self.width, h = top, dx = 0, dy = 0 }
	for i = 0, self.height - top, floor((self.height - top)/4) do
		self.steps[#self.steps+1] = {
			o = { x = 0, y = top + i, w = self.width, h = self.height - top - i, dx = 0, dy = top             },
			n = { x = 0, y = top,     w = self.width, h = i,                     dx = 0, dy = self.height - i },
		}
	end
end


function pushUpV(self)
	local top = self.ext * 2
	self.steps  = {}
	self.static = { x = top, y = 0, w = self.width - top, h = self.height, dx = top, dy = 0 }
	for i = 0, top, floor((self.width - top)/4) do
		self.steps[#self.steps+1] = {
			o = { x = 0,       y = 0, w = top - i, h = self.height, dx = i, dy = 0 },
			n = { x = top - i, y = 0, w = i,       h = self.height, dx = 0, dy = 0 },
		}
	end
end


function pushDownH(self)
	local top = self.height - self.ext * 2
	self.steps  = {}
	self.static = { x = 0, y = 0, w = self.width, h = top, dx = 0, dy = 0 }
	for i = 0, self.height - top, floor((self.height - top)/4) do
		self.steps[#self.steps+1] = {
			o = { x = 0, y = top,             w = self.width, h = self.height - top - i, dx = 0, dy = top + i },
			n = { x = 0, y = self.height - i, w = self.width, h = i,                     dx = 0, dy = top     },
		}
	end
end


function pushDownV(self)
	local top = self.ext * 2
	self.steps  = {}
	self.static = { x = top, y = 0, w = self.width - top, h = self.height, dx = top, dy = 0 }
	for i = 0, top, floor((self.width - top)/4) do
		self.steps[#self.steps+1] = {
			o = { x = i, y = 0, w = top - i, h = self.height, dx = 0      , dy = 0 },
			n = { x = 0, y = 0, w = i,       h = self.height, dx = top - i, dy = 0 },
		}
	end
end


function bumpDownH(self)
	local top = self.height - self.ext * 2
	self.static = { x = 0, y = 0, w = self.width, h = top, dx = 0, dy = 0 }
	self.steps  = {
		{ n = { x = 0, y = top, w = self.width, h = self.height - top - 4, dx = 0, dy = top + 4  },
		  o = { x = 0, y = self.height - 4, w = self.width, h = 4,         dx = 0, dy = top      }, },
		{ n = { x = 0, y = 0,   w = self.width, h = self.height,           dx = 0, dy = 0        }, },
	}
end


function bumpDownV(self)
	local top = self.ext * 2
	self.static = { x = top, y = 0, w = self.width, h = self.height - top, dx = top, dy = 0 }
	self.steps  = {
		{ n = { x = 4, y = 0, w = top - 4,    h = self.height, dx = 0,       dy = 0 },
		  o = { x = 0, y = 0, w = 4,          h = self.height, dx = top - 4, dy = 0 }, },
		{ n = { x = 0, y = 0, w = self.width, h = self.height, dx = 0,       dy = 0 }, },
	}
end


function bumpUpH(self)
	local top = self.height - self.ext * 2
	self.static = { x = 0, y = 0, w = self.width, h = top, dx = 0, dy = 0 }
	self.steps  = {
		{ n = { x = 0, y = top + 4, w = self.width, h = self.height - top - 4, dx = 0, dy = top      },
		  o = { x = 0, y = top,     w = self.width, h = 4,             dx = 0, dy = self.height - 4  }, },
		{ n = { x = 0, y = 0,       w = self.width, h = self.height,           dx = 0, dy = 0        }, },
	}
end


function bumpUpV(self)
	local top = self.ext * 2
	self.static = { x = top, y = 0, w = self.width, h = self.height - top, dx = top, dy = 0 }
	self.steps  = {
		{ n = { x = 0,       y = 0, w = top - 4,    h = self.height, dx = 4, dy = 0 },
		  o = { x = top - 4, y = 0, w = 4,          h = self.height, dx = 0, dy = 0 }, },
		{ n = { x = 0,       y = 0, w = self.width, h = self.height, dx = 0, dy = 0 }, },
	}
end


function displayName(self)
	return "RetroBrowser"
end


function skin(self, s)
	s.retro = {
		layout = Window.noLayout
	}
end


function free(self)
	log:debug("RetroBrowser:free()")

	-- reset subscription to showbriefly and unsubscribe
	self.player.slimServer.comet:subscribe(
		self.subid,
		self.oldcallback,
		self.player.id,
		{ 'displaystatus', 'subscribe:showbriefly' }
	)

	self.player.slimServer.comet:unsubscribe(self.subid, self.callback)
	
	return true
end

